""" Tests of DataFetcher functionality """
import logging
import os
from pathlib import Path

# pylint: disable=E0401, E0402, R0903, R1721, W0611, W0613, W0621

from typing import List
from unittest.mock import patch, MagicMock

import pytest
from datafetcher import locations_report
from datafetcher.locations_report import NgasTarget
from datafetcher.retrieve_launcher import ParallelRetrieverLauncher
from datafetcher.datafetcher import DataFetcher
from datafetcher.errors import (
    NGASFetchError,
    NoLocatorException,
    LocationServiceTimeoutException,
    NGASServiceRedirectsException,
    NGASServiceErrorException,
)

from datafetcher.ngas_retriever import NgasNonRetriever, NgasRetriever
from datafetcher.utilities import (
    RetrievalMode,
    CLIParam,
    LOG_FORMAT,
    LUSTRE,
    ASDM_BIN_SUBDIR,
)

from .conftest import (
    CAL_LOCATOR,
    mock_retrieve_sdms,
    mock_retrieve_exec_block,
    vla_small_eb_locations_report,
    eb_17a_109_18468_locations_report,
    eb_17a_109_18468_locations_file,
)

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)

_RUN_ALL = True


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_fetches_cal_from_locator(make_tempdir, cal_locations_report):
    """
    Test vanilla fetch using product locator

    :param make_tempdir:
    :return:
    """
    output_dir = Path(make_tempdir)
    os.chdir(output_dir)
    print(f"output will be written to {output_dir}")

    args = [
        CLIParam.SPL.value,
        CAL_LOCATOR,
    ]

    with patch(
        "datafetcher.locations_report.LocationsReportBuilder.build",
        return_value=cal_locations_report,
    ):
        datafetcher = DataFetcher(args)

    loc_report = datafetcher.locations_report
    assert loc_report.aggregate_size == 27668480

    target = loc_report.targets[0]
    assert target.subdirectory == "18B-265_2019_12_10_T00_00_59.203.tar"
    assert target.relative_path == target.subdirectory

    datafetcher.run = MagicMock(name="run")
    datafetcher.run()
    datafetcher.run.assert_called_once()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_fetches_cal_from_file(cal_json_file):
    """
    Test vanilla fetch using .json locations report file.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file)]
    datafetcher = DataFetcher(args)

    loc_report = datafetcher.locations_report
    assert len(loc_report.targets) == 1
    target = loc_report.targets[0]
    assert target.checksum == "722663767"

    datafetcher.run = MagicMock(name="run")
    datafetcher.run()
    datafetcher.run.assert_called_once()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_retrieves_sdms(
    make_tempdir, eb_17a_109_18468_locations_file, eb_17a_109_18468_locations_report
):
    """
    If dry_run is not set, expected number of files should be written,
    with expected sizes.

    :return:
    """
    args = [CLIParam.FILE.value, str(eb_17a_109_18468_locations_file)]
    datafetcher = DataFetcher(args)
    with patch(
        "datafetcher.ngas_retriever.NgasRetriever.retrieve",
        return_value=mock_retrieve_sdms(
            make_tempdir, eb_17a_109_18468_locations_report.targets, force=False
        ),
    ):
        datafetcher.run()
        written = [file for file in Path(make_tempdir).iterdir()]
        assert len(written) == 36


class NonFetchingNgasCopyLauncher(ParallelRetrieverLauncher):
    """ A fake ParallelNgasAdapter that doesn't retrieve anything """

    def launch_retrievers(self) -> int:
        """
        Launches fake retrievers that pretend to retrieve products.

        :return:
        """

        retriever = NgasNonRetriever(self.ngas_targets, force_overwrite=False)
        return retriever.retrieve()


class LustreDsocDataFetcher(DataFetcher):
    """ This is a DataFetcher executing on /lustre at the DSOC. """

    def __init__(self, make_tempdir, args_in: List[str]):
        """
        We're mocking DataFetcher.__init__ so we can set the current/output
        directory to one containing "/lustre".

        :param args_in: command-line parameters

        """
        super().__init__(args_in)

        self.output_dir = Path(make_tempdir) / LUSTRE
        self.output_dir.mkdir()
        os.chdir(self.output_dir)

        assert f"/{LUSTRE}" in str(Path.cwd())

    def run(self):
        """
        Pretend to fetch the product(s).

        :return:
        """
        launcher = NonFetchingNgasCopyLauncher(
            self.locations_report.targets, dry_run=False, force_overwrite=False
        )

        count = 0
        expected = len(self.locations_report.targets)
        for target in self.locations_report.targets:
            count += 1
            if not self.dry_run:
                if not launcher.launch_retrievers():
                    raise NGASFetchError(f"{target.relative_path}")

        if count < expected:
            raise NGASFetchError(f"{expected} files expected, but only {count} retrieved")


def mock_determine_retrieve_mode(*args) -> RetrievalMode:
    """
    Force an NGAS direct copy

    :param args: included only to fulfill contract
    :return:
    """
    return RetrievalMode.COPY


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_can_direct_copy(monkeypatch, make_tempdir, cal_json_file):
    """
    Can we execute a direct-copy NGAS request?

    :param monkeypatch:
    :param make_tempdir:
    :param cal_json_file:
    :return:
    """

    monkeypatch.setattr(
        locations_report.NgasTarget,
        "determine_retrieve_mode",
        mock_determine_retrieve_mode,
    )

    with patch("datafetcher.ngas_retriever.DryCopier.send_request") as mock_copy:
        with patch("datafetcher.ngas_retriever.check_result"):
            with patch("datafetcher.ngas_retriever.DryStreamer.send_request") as mock_stream:
                args = [CLIParam.FILE.value, str(cal_json_file)]
                datafetcher = LustreDsocDataFetcher(make_tempdir, args)
                datafetcher.run()
                mock_copy.assert_called()
                mock_stream.assert_not_called()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_direct_copy_lustre_only(monkeypatch, make_tempdir, cal_json_file):
    """
    Trying to pull files out of NGAS via the direct-copy plugin
    works only on /lustre, and then only if execution site is DSOC.

    :return:
    """

    args = [CLIParam.FILE.value, str(cal_json_file)]
    datafetcher = DataFetcher(args)
    loc_report = datafetcher.locations_report
    target = loc_report.targets[0]
    assert target.retrieve_method != RetrievalMode.COPY

    monkeypatch.setattr(
        NgasTarget,
        "determine_retrieve_mode",
        mock_determine_retrieve_mode,
    )
    datafetcher = LustreDsocDataFetcher(make_tempdir, args)
    loc_report = datafetcher.locations_report
    target = loc_report.targets[0]
    assert target.retrieve_method == RetrievalMode.COPY


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_bad_location_file_throws_file_not_found():
    """
    If it can't find the locations file, it should say so

    :return:
    """
    args = [CLIParam.FILE.value, "move_along_nothing_to_see_here"]
    with pytest.raises(FileNotFoundError):
        DataFetcher(args)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_bad_locator_throws_no_locator():
    """
    If the locator's no good, it should say so

    :return:
    """
    args = [CLIParam.SPL.value, "this will not end well"]
    with pytest.raises(NoLocatorException):
        DataFetcher(args)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_service_timeout_throws_expected(monkeypatch):
    """
    If the locator service times out, please tell us upon crashing.

    :return:
    """

    def mock_timeout(*args, **kwargs):
        raise LocationServiceTimeoutException()

    args = [CLIParam.SPL.value, CAL_LOCATOR]
    monkeypatch.setattr(NgasRetriever, "retrieve", mock_timeout)
    with pytest.raises(LocationServiceTimeoutException):
        DataFetcher(args).run()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_service_too_many_redirects_throws_expected(monkeypatch):
    """
    If the locator service exceeds the maximum number of redirects,
    please tell us upon crashing.

    :return:
    """

    def mock_too_many_redirects(*args, **kwargs):
        raise NGASServiceRedirectsException()

    args = [CLIParam.SPL.value, CAL_LOCATOR]
    monkeypatch.setattr(NgasRetriever, "retrieve", mock_too_many_redirects)
    with pytest.raises(NGASServiceRedirectsException):
        DataFetcher(args).run()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_catastrophic_service_error_throws_expected(monkeypatch):
    """
    If the locator service throws a catastrophic error,
    please tell us upon crashing.

    :return:
    """

    def mock_catastrophe(*args, **kwargs):
        raise NGASFetchError()

    args = [CLIParam.SPL.value, CAL_LOCATOR]
    monkeypatch.setattr(NgasRetriever, "retrieve", mock_catastrophe)
    with pytest.raises(NGASFetchError):
        DataFetcher(args).run()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_throws_service_error_with_bad_server_info(bad_server_json_file):
    """
    If we give it an invalid server, it should complain about that specifically.

    :return:
    """

    datafetcher = DataFetcher([CLIParam.FILE.value, str(bad_server_json_file)])
    with pytest.raises(NGASServiceErrorException):
        datafetcher.run()


# @pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_makes_data_fetcher_from_file(make_tempdir, vla_small_locations_file):
    """
    Will datafetcher accept a locations file?

    :return:
    """

    os.chdir(Path(make_tempdir))
    args = [CLIParam.FILE.value, str(vla_small_locations_file)]
    datafetcher = DataFetcher(args)
    with patch("time.sleep"):
        with patch("datafetcher.ngas_retriever.NgasCopier.send_request", return_value=1):
            with patch("datafetcher.ngas_retriever.NgasStreamer.send_request", return_value=1):
                retrieved_count = datafetcher.run()
                assert retrieved_count == 44


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_makes_data_fetcher_from_service():
    """
    Will datafetcher accept a product locator?

    :return:
    """
    args = [CLIParam.SPL.value, CAL_LOCATOR]
    datafetcher = DataFetcher(args)
    with patch("datafetcher.ngas_retriever.NgasRetriever.retrieve", return_value=1):
        retrieved_count = datafetcher.run()
        assert retrieved_count == 1


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_output_paths_are_unique(make_tempdir, vla_small_locations_file):
    """
    When retrieving multiple files, be sure each destination is unique.

    :param make_tempdir: top-level temporary directory
    :return:
    """

    os.chdir(Path(make_tempdir))
    args = [CLIParam.FILE.value, str(vla_small_locations_file)]
    datafetcher = DataFetcher(args)
    expected_file_count = len(datafetcher.locations_report.targets)
    with patch("time.sleep"):
        with patch(
            "datafetcher.ngas_retriever.NgasCopier.send_request", return_value=1
        ) as mock_copy:
            with patch(
                "datafetcher.ngas_retriever.NgasStreamer.send_request",
                return_value=1,
            ) as mock_stream:
                datafetcher.run()
                assert mock_stream.call_count + mock_copy.call_count == expected_file_count


def test_eb_sdms_and_bdfs_written_to_correct_subdirs(
    make_tempdir, vla_small_locations_file, vla_small_eb_locations_report
):
    """
    The SDMs in an execution block should be written to the output directory;
    the BDFs should be written to output_dir/ASDMBinary/

    :return:
    """
    args = [CLIParam.FILE.value, str(vla_small_locations_file)]
    datafetcher = DataFetcher(args)
    with patch(
        "datafetcher.ngas_retriever.NgasRetriever.retrieve",
        return_value=mock_retrieve_exec_block(
            make_tempdir, vla_small_eb_locations_report.targets, force=False
        ),
    ):

        datafetcher.run()
        output_dir = Path(make_tempdir)
        written = [file for file in output_dir.rglob("*") if file.is_file()]
        assert len(written) == 44

        for file in written:
            if file.is_file():
                if file.name.endswith(".bdf"):
                    assert file.parent.name == ASDM_BIN_SUBDIR
                else:
                    assert file.parent.name == output_dir.name
